from .base import ModuleTestBase


class TestChaos(ModuleTestBase):
    config_overrides = {"modules": {"chaos": {"api_key": "asdf"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://dns.projectdiscovery.io/dns/example.com",
            match_headers={"Authorization": "asdf"},
            json={"domain": "example.com", "subdomains": 65},
        )
        module_test.httpx_mock.add_response(
            url="https://dns.projectdiscovery.io/dns/blacklanternsecurity.com/subdomains",
            match_headers={"Authorization": "asdf"},
            json={
                "domain": "blacklanternsecurity.com",
                "subdomains": [
                    "*.asdf.cloud",
                ],
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.cloud.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from .base import ModuleTestBase


class TestCRT_DB(ModuleTestBase):
    async def setup_after_prep(self, module_test):
        class AsyncMock:
            async def fetch(self, *args, **kwargs):
                return [
                    {"name_value": "asdf.blacklanternsecurity.com"},
                    {"name_value": "zzzz.blacklanternsecurity.com"},
                ]

            async def close(self):
                pass

        async def mock_connect(*args, **kwargs):
            return AsyncMock()

        module_test.monkeypatch.setattr("asyncpg.connect", mock_connect)

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from bbot.modules.output.base import BaseOutputModule
import markdown
import html


class web_report(BaseOutputModule):
    watched_events = ["URL", "TECHNOLOGY", "FINDING", "VULNERABILITY", "VHOST"]
    meta = {
        "description": "Create a markdown report with web assets",
        "created_date": "2023-02-08",
        "author": "@liquidsec",
    }
    options = {
        "output_file": "",
        "css_theme_file": "https://cdnjs.cloudflare.com/ajax/libs/github-markdown-css/5.1.0/github-markdown.min.css",
    }
    options_desc = {"output_file": "Output to file", "css_theme_file": "CSS theme URL for HTML output"}
    deps_pip = ["markdown~=3.4.3"]

    async def setup(self):
        html_css_file = self.config.get("css_theme_file", "")

        self.html_header = f"""
        <!DOCTYPE html>
        <html>
        <head>
        <link rel="stylesheet" href="{html_css_file}">
        </head>
        <body>
        """

        self.html_footer = "</body></html>"
        self.web_assets = {}
        self.markdown = ""

        self._prep_output_dir("web_report.html")
        return True

    async def handle_event(self, event):
        if event.type == "URL":
            parsed = event.parsed_url
            host = f"{parsed.scheme}://{parsed.netloc}/"
            if host not in self.web_assets.keys():
                self.web_assets[host] = {"URL": []}
            parent_chain = []

            current_parent = event.parent
            while not current_parent.type == "SCAN":
                parent_chain.append(
                    f" ({current_parent.module})---> [{current_parent.type}]:{html.escape(current_parent.pretty_string)}"
                )
                current_parent = current_parent.parent

            parent_chain.reverse()
            parent_chain_text = (
                "".join(parent_chain)
                + f" ({event.module})---> "
                + f"[{event.type}]:{html.escape(event.pretty_string)}"
            )
            self.web_assets[host]["URL"].append(f"**{html.escape(event.data)}**: {parent_chain_text}")

        else:
            current_parent = event.parent
            parsed = None
            while 1:
                if current_parent.type == "URL":
                    parsed = current_parent.parsed_url
                    break
                current_parent = current_parent.parent
                if current_parent.parent.type == "SCAN":
                    break
            if parsed:
                host = f"{parsed.scheme}://{parsed.netloc}/"
                if host not in self.web_assets.keys():
                    self.web_assets[host] = {"URL": []}
                if event.type not in self.web_assets[host].keys():
                    self.web_assets[host][event.type] = [html.escape(event.pretty_string)]
                else:
                    self.web_assets[host][event.type].append(html.escape(event.pretty_string))

    async def report(self):
        for host in self.web_assets.keys():
            self.markdown += f"# {host}\n\n"

            for event_type in self.web_assets[host].keys():
                self.markdown += f"### {event_type}\n"
                dedupe = []
                for e in self.web_assets[host][event_type]:
                    if e in dedupe:
                        continue
                    dedupe.append(e)
                    self.markdown += f"\n* {e}\n"
                self.markdown += "\n"

        if self.file is not None:
            self.file.write(self.html_header)
            self.file.write(markdown.markdown(self.markdown))
            self.file.write(self.html_footer)
            self.file.flush()
            self.info(f"Web Report saved to {self.output_file}")



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class BufferOverrun(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query BufferOverrun's TLS API for subdomains",
        "created_date": "2024-10-23",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": "", "commercial": False}
    options_desc = {"api_key": "BufferOverrun API key", "commercial": "Use commercial API"}

    base_url = "https://tls.bufferover.run/dns"
    commercial_base_url = "https://bufferover-run-tls.p.rapidapi.com/ipv4/dns"

    async def setup(self):
        self.commercial = self.config.get("commercial", False)
        return await super().setup()

    def prepare_api_request(self, url, kwargs):
        if self.commercial:
            kwargs["headers"]["x-rapidapi-host"] = "bufferover-run-tls.p.rapidapi.com"
            kwargs["headers"]["x-rapidapi-key"] = self.api_key
        else:
            kwargs["headers"]["x-api-key"] = self.api_key
        return url, kwargs

    async def request_url(self, query):
        url = f"{self.commercial_base_url if self.commercial else self.base_url}?q=.{query}"
        return await self.api_request(url)

    async def parse_results(self, r, query):
        j = r.json()
        subdomains_set = set()
        if isinstance(j, dict):
            results = j.get("Results", [])
            for result in results:
                parts = result.split(",")
                if len(parts) > 4:
                    subdomain = parts[4].strip()
                    if subdomain and subdomain.endswith(f".{query}"):
                        subdomains_set.add(subdomain)
        return subdomains_set



import ipaddress
import regex as re
from functools import cached_property
from bbot.errors import ValidationError
from bbot.core.helpers import validators
from bbot.core.helpers.misc import split_host_port, make_ip_type
from bbot.core.helpers import regexes, smart_decode, smart_encode_punycode

bbot_event_seeds = {}


"""
An "Event Seed" is a lightweight event containing only the minimum logic required to:
    - parse input to determine the event type + data
    - validate+sanitize the data
    - extract the host for scope purposes

It's useful for quickly parsing target lists without the cpu+memory overhead of creating full-fledged BBOT events

Not every type of BBOT event needs to be represented here. Only ones that are meant to be targets.
"""


class EventSeedRegistry(type):
    """
    Metaclass for EventSeed that registers all subclasses in a registry.
    """

    def __new__(mcs, name, bases, attrs):
        global bbot_event_seeds
        cls = super().__new__(mcs, name, bases, attrs)
        # Don't register the base EventSeed class
        if name != "BaseEventSeed":
            bbot_event_seeds[cls.__name__] = cls
        return cls


def EventSeed(input):
    input = smart_encode_punycode(smart_decode(input).strip())
    for _, event_class in bbot_event_seeds.items():
        if hasattr(event_class, "precheck"):
            if event_class.precheck(input):
                return event_class(input)
        else:
            for regex in event_class.regexes:
                match = regex.match(input)
                if match:
                    data = event_class.handle_match(match)
                    return event_class(data)
    raise ValidationError(f'Unable to autodetect data type from "{input}"')


class BaseEventSeed(metaclass=EventSeedRegistry):
    regexes = []
    _target_type = "TARGET"

    __slots__ = ["data", "host", "port", "input"]

    def __init__(self, data):
        self.data, self.host, self.port = self._sanitize_and_extract_host(data)
        self.input = self._override_input(data)

    @staticmethod
    def handle_match(match):
        """
        Given a regex match, returns the event data
        """
        return match.group(0)

    def _sanitize_and_extract_host(self, data):
        """
        Given the event data, returns the host

        Returns:
            tuple: (data, host, port)
        """
        return data, None, None

    def _override_input(self, input):
        return self.data

    @property
    def type(self):
        return self.__class__.__name__

    @cached_property
    def _hash(self):
        return hash(self.input)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __str__(self):
        return f"EventSeed({self.input})"

    def __repr__(self):
        return str(self)


class IP_ADDRESS(BaseEventSeed):
    regexes = regexes.event_type_regexes["IP_ADDRESS"]

    @staticmethod
    def precheck(data):
        try:
            return ipaddress.ip_address(data)
        except ValueError:
            return False

    @staticmethod
    def _sanitize_and_extract_host(data):
        validated = ipaddress.ip_address(data)
        return str(validated), validated, None


class DNS_NAME(BaseEventSeed):
    regexes = regexes.event_type_regexes["DNS_NAME"]

    @staticmethod
    def _sanitize_and_extract_host(data):
        validated = validators.validate_host(data)
        return validated, validated, None


class IP_RANGE(BaseEventSeed):
    regexes = regexes.event_type_regexes["IP_RANGE"]

    @staticmethod
    def precheck(data):
        try:
            return ipaddress.ip_network(str(data), strict=False)
        except ValueError:
            return False

    @staticmethod
    def _sanitize_and_extract_host(data):
        validated = ipaddress.ip_network(str(data), strict=False)
        return str(validated), validated, None


class OPEN_TCP_PORT(BaseEventSeed):
    regexes = regexes.event_type_regexes["OPEN_TCP_PORT"]

    @staticmethod
    def _sanitize_and_extract_host(data):
        validated = validators.validate_open_port(data)
        host, port = split_host_port(validated)
        host = make_ip_type(host)
        return str(validated), host, port


class URL_UNVERIFIED(BaseEventSeed):
    regexes = regexes.event_type_regexes["URL"]

    _scheme_to_port = {
        "https": 443,
        "http": 80,
    }

    @staticmethod
    def _sanitize_and_extract_host(data):
        parsed_url = validators.clean_url(data, url_querystring_remove=False)
        scheme = parsed_url.scheme
        host = make_ip_type(validators.validate_host(parsed_url.hostname))
        port = parsed_url.port
        if port is None:
            port = URL_UNVERIFIED._scheme_to_port.get(scheme, None)
        return parsed_url.geturl(), host, port


class EMAIL_ADDRESS(BaseEventSeed):
    regexes = regexes.event_type_regexes["EMAIL_ADDRESS"]

    @staticmethod
    def _sanitize_and_extract_host(data):
        validated = validators.validate_email(data)
        host = validated.rsplit("@", 1)[-1]
        host, port = split_host_port(host)
        return validated, host, port


class ORG_STUB(BaseEventSeed):
    regexes = (re.compile(r"^(?:ORG|ORG_STUB):(.*)"),)

    def _override_input(self, input):
        return f"ORG_STUB:{self.data}"

    @staticmethod
    def handle_match(match):
        return match.group(1)


class USERNAME(BaseEventSeed):
    regexes = (re.compile(r"^(?:USER|USERNAME):(.*)"),)

    def _override_input(self, input):
        return f"USERNAME:{self.data}"

    @staticmethod
    def handle_match(match):
        return match.group(1)


class FILESYSTEM(BaseEventSeed):
    regexes = (re.compile(r"^(?:FILESYSTEM|FILE|FOLDER|DIR|PATH):(.*)"),)

    def _override_input(self, input):
        return f"FILESYSTEM:{self.data['path']}"

    @staticmethod
    def handle_match(match):
        return {"path": match.group(1)}


class MOBILE_APP(BaseEventSeed):
    regexes = (re.compile(r"^(?:MOBILE_APP|APK|IPA|APP):(.*)"),)

    def _override_input(self, input):
        return f"MOBILE_APP:{self.data['url']}"

    @staticmethod
    def handle_match(match):
        return {"url": match.group(1)}


class BLACKLIST_REGEX(BaseEventSeed):
    regexes = (re.compile(r"^(?:RE|REGEX):(.*)"),)
    _target_type = "BLACKLIST"

    def _override_input(self, input):
        return f"REGEX:{self.data}"

    @staticmethod
    def handle_match(match):
        return match.group(1)



import re

from bbot.modules.templates.subdomain_enum import subdomain_enum


class dnsdumpster(subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query dnsdumpster for subdomains",
        "created_date": "2022-03-12",
        "author": "@TheTechromancer",
    }

    base_url = "https://dnsdumpster.com"

    async def query(self, domain):
        ret = []
        # first, get the CSRF tokens
        res1 = await self.api_request(self.base_url)
        status_code = getattr(res1, "status_code", 0)
        if status_code in [429]:
            self.verbose(f'Too many requests "{status_code}"')
            return ret
        elif status_code not in [200]:
            self.verbose(f'Bad response code "{status_code}" from DNSDumpster')
            return ret
        else:
            self.debug(f'Valid response code "{status_code}" from DNSDumpster')

        html = self.helpers.beautifulsoup(res1.content, "html.parser")
        if html is False:
            self.verbose("BeautifulSoup returned False")
            return ret

        csrftoken = None
        csrfmiddlewaretoken = None
        try:
            for cookie in res1.headers.get("set-cookie", "").split(";"):
                try:
                    k, v = cookie.split("=", 1)
                except ValueError:
                    self.verbose("Error retrieving cookie")
                    return ret
                if k == "csrftoken":
                    csrftoken = str(v)
            csrfmiddlewaretoken = html.find("input", {"name": "csrfmiddlewaretoken"}).attrs.get("value", None)
        except AttributeError:
            pass

        # Abort if we didn't get the tokens
        if not csrftoken or not csrfmiddlewaretoken:
            self.verbose("Error obtaining CSRF tokens")
            self.errorState = True
            return ret
        else:
            self.debug("Successfully obtained CSRF tokens")

        if self.scan.stopping:
            return

        # Otherwise, do the needful
        subdomains = set()
        res2 = await self.api_request(
            f"{self.base_url}/",
            method="POST",
            cookies={"csrftoken": csrftoken},
            data={
                "csrfmiddlewaretoken": csrfmiddlewaretoken,
                "targetip": str(domain).lower(),
                "user": "free",
            },
            headers={
                "origin": "https://dnsdumpster.com",
                "referer": "https://dnsdumpster.com/",
            },
        )
        status_code = getattr(res2, "status_code", 0)
        if status_code not in [200]:
            self.verbose(f'Bad response code "{status_code}" from DNSDumpster')
            return ret
        html = self.helpers.beautifulsoup(res2.content, "html.parser")
        if html is False:
            self.verbose("BeautifulSoup returned False")
            return ret
        escaped_domain = re.escape(domain)
        match_pattern = re.compile(r"^[\w\.-]+\." + escaped_domain + r"$")
        for subdomain in html.findAll(text=match_pattern):
            subdomains.add(str(subdomain).strip().lower())

        return list(subdomains)



import base64
import hashlib
from .base import BaseLightfuzz
from bbot.errors import HttpCompareError
from urllib.parse import unquote, quote


# Global cache for compiled YARA rules
_compiled_rules_cache = None


class crypto(BaseLightfuzz):
    """
    Detects the use of cryptography in web parameters, and probes for some cryptographic vulnerabilities

    * Cryptographic Error Detection:
       - Detects known cryptographic error messages in server responses.

    * Cryptographic Parameter Value Detection:
       - Detects use of cryptography in web parameter values.
       - Validates by attempting to manipulate the value regardless of its encoding.

    * Length Extension Attack Detection:
       - Identifies parameters which may be expecting hash digests for values, and any linked parameters which invalidate them.

    * Padding Oracle Vulnerabilities:
       - Identifies the presence of cryptographic oracles that could be exploited to arbitrary decrypt or encrypt data for the parameter value.


    """

    friendly_name = "Cryptography Probe"

    # Although we have an envelope system to detect hex and base64 encoded parameter values, those are only assigned when they decode to a valid string.
    # Since crypto values (and serialized objects) will not decode properly, we need a more concise check here to determine how to process them.

    @staticmethod
    def is_hex(s):
        try:
            bytes.fromhex(s)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_base64(s):
        try:
            if base64.b64encode(base64.b64decode(s)).decode() == s:
                return True
        except Exception:
            return False
        return False

    # A list of YARA rules for detecting cryptographic error messages
    crypto_error_strings = [
        "invalid mac",
        "padding is invalid",
        "bad data",
        "length of the data to decrypt is invalid",
        "specify a valid key size",
        "invalid algorithm specified",
        "object already exists",
        "key does not exist",
        "the parameter is incorrect",
        "cryptography exception",
        "access denied",
        "unknown error",
        "invalid provider type",
        "no valid cert found",
        "cannot find the original signer",
        "signature description could not be created",
        "crypto operation failed",
        "OpenSSL Error",
    ]

    @property
    def compiled_rules(self):
        """
        We need to cache the compiled YARA rule globally since lightfuzz submodules are recreated for every handle_event
        """
        global _compiled_rules_cache
        if _compiled_rules_cache is None:
            _compiled_rules_cache = self.lightfuzz.helpers.yara.compile_strings(self.crypto_error_strings, nocase=True)
        return _compiled_rules_cache

    @staticmethod
    def format_agnostic_decode(input_string, urldecode=False):
        """
        Decodes a string from either hex or base64 (without knowing which first), and optionally URL-decoding it first.

        Parameters:
        - input_string (str): The string to decode.
        - urldecode (bool): If True, URL-decodes the input first.

        Returns:
        - tuple: (decoded data, encoding type: 'hex', 'base64', or 'unknown').
        """
        encoding = "unknown"
        if urldecode:
            input_string = unquote(input_string)
        if BaseLightfuzz.is_hex(input_string):
            data = bytes.fromhex(input_string)
            encoding = "hex"
        elif BaseLightfuzz.is_base64(input_string):
            data = base64.b64decode(input_string)
            encoding = "base64"
        else:
            data = str
        return data, encoding

    @staticmethod
    def format_agnostic_encode(data, encoding, urlencode=False):
        """
        Encodes data into hex or base64, with optional URL-encoding.

        Parameters:
        - data (bytes): The data to encode.
        - encoding (str): The encoding type ('hex' or 'base64').
        - urlencode (bool): If True, URL-encodes the result.

        Returns:
        - str: The encoded data as a string.

        Raises:
        - ValueError: If an unsupported encoding type is specified.
        """
        if encoding == "hex":
            encoded_data = data.hex()
        elif encoding == "base64":
            encoded_data = base64.b64encode(data).decode("utf-8")  # base64 encoding returns bytes, decode to string
        else:
            raise ValueError("Unsupported encoding type specified")
        if urlencode:
            return quote(encoded_data)
        return encoded_data

    @staticmethod
    def modify_string(input_string, action="truncate", position=None, extension_length=1):
        """
        Modifies a cryptographic string by either truncating it, mutating a byte at a specified position, or extending it with null bytes.

        Parameters:
        - input_string (str): The string to modify.
        - action (str): The action to perform ('truncate', 'mutate', 'extend').
        - position (int): The position to mutate (only used if action is 'mutate').
        - extension_length (int): The number of null bytes to add if action is 'extend'.

        Returns:
        - str: The modified string.
        """
        if not isinstance(input_string, str):
            input_string = str(input_string)

        data, encoding = crypto.format_agnostic_decode(input_string)
        if encoding != "base64" and encoding != "hex":
            raise ValueError("Input must be either hex or base64 encoded")

        if action == "truncate":
            modified_data = data[:-1]  # Remove the last byte
        elif action == "mutate":
            if not position:
                position = len(data) // 2
            if position < 0 or position >= len(data):
                raise ValueError("Position out of range")
            byte_list = list(data)
            byte_list[position] = (byte_list[position] + 1) % 256
            modified_data = bytes(byte_list)
        elif action == "extend":
            modified_data = data + (b"\x00" * extension_length)
        elif action == "flip":
            if not position:
                position = len(data) // 2
            if position < 0 or position >= len(data):
                raise ValueError("Position out of range")
            byte_list = list(data)
            byte_list[position] ^= 0xFF  # Flip all bits in the byte at the specified position
            modified_data = bytes(byte_list)
        else:
            raise ValueError("Unsupported action")
        return crypto.format_agnostic_encode(modified_data, encoding)

    # Check if the entropy of the data is greater than the threshold, indicating it is likely encrypted
    def is_likely_encrypted(self, data, threshold=4.5):
        entropy = self.lightfuzz.helpers.calculate_entropy(data)
        return entropy >= threshold

    # Perform basic cryptanalysis on the input string, attempting to determine if it is likely encrypted and if it is a block cipher
    def cryptanalysis(self, input_string):
        likely_crypto = False
        possible_block_cipher = False
        data, encoding = self.format_agnostic_decode(input_string)
        likely_crypto = self.is_likely_encrypted(data)
        data_length = len(data)
        if data_length % 8 == 0:
            possible_block_cipher = True
        return likely_crypto, possible_block_cipher

    # Determine possible block sizes for a given ciphertext length
    @staticmethod
    def possible_block_sizes(ciphertext_length):
        potential_block_sizes = [8, 16]
        possible_sizes = []
        for block_size in potential_block_sizes:
            num_blocks = ciphertext_length // block_size
            if ciphertext_length % block_size == 0 and num_blocks >= 2:
                possible_sizes.append(block_size)
        return possible_sizes

    async def padding_oracle_execute(self, original_data, encoding, block_size, cookies, possible_first_byte=True):
        """
        Execute the padding oracle attack for a given block size.
        The goal here is not actual exploitation (arbitrary encryption or decryption), but rather to definitively confirm whether padding oracle vulnerability exists and is exploitable.

        Parameters:
        - original_data (bytes): The original ciphertext data.
        - encoding (str): The encoding type ('hex' or 'base64').
        - block_size (int): The block size to use for the padding oracle attack.
        - cookies (dict): Cookies to include, if any
        - possible_first_byte (bool): If True, use the first byte as the baseline byte.

        Returns:
        - bool: True if the padding oracle attack is successful.
        """
        ivblock = b"\x00" * block_size  # initialize the IV block with null bytes
        paddingblock = b"\x00" * block_size  # initialize the padding block with null bytes
        datablock = original_data[-block_size:]  # extract the last block of the original data

        # This handling the 1/255 chance that the first byte is correct padding which would cause a false negative.
        if possible_first_byte:
            baseline_byte = b"\xff"  # set the baseline byte to 0xff
            starting_pos = 0  # set the starting position to 0
        else:
            baseline_byte = b"\x00"  # set the baseline byte to 0x00
            starting_pos = 1  # set the starting position to 1
        # first obtain
        baseline = self.compare_baseline(
            self.event.data["type"],
            self.format_agnostic_encode(ivblock + paddingblock[:-1] + baseline_byte + datablock, encoding),
            cookies,
        )
        differ_count = 0
        # for each possible byte value, send a probe and check if the response is different
        for i in range(starting_pos, starting_pos + 254):
            byte = bytes([i])
            oracle_probe = await self.compare_probe(
                baseline,
                self.event.data["type"],
                self.format_agnostic_encode(ivblock + paddingblock[:-1] + byte + datablock, encoding),
                cookies,
            )
            # oracle_probe[0] will be false if the response is different - oracle_probe[1] stores what aspect of the response is different (headers, body, code)
            if oracle_probe[0] is False and "body" in oracle_probe[1]:
                differ_count += 1

                if i == 2:
                    if possible_first_byte is True:
                        # Thats two results which appear "different". Since this is the first run, it's entirely possible \x00 was the correct padding.
                        # We will break from this loop and redo it with the last byte as the baseline instead of the first
                        return None
                    else:
                        # Now that we have tried the run twice, we know it can't be because the first byte was the correct padding, and we know it is not vulnerable
                        return False
        # A padding oracle vulnerability will produce exactly one different response, and no more, so this is likely a real padding oracle
        if differ_count == 1:
            return True
        return False

    async def padding_oracle(self, probe_value, cookies):
        data, encoding = self.format_agnostic_decode(probe_value)
        possible_block_sizes = self.possible_block_sizes(
            len(data)
        )  # determine possible block sizes for the ciphertext

        for block_size in possible_block_sizes:
            padding_oracle_result = await self.padding_oracle_execute(data, encoding, block_size, cookies)
            # if we get a negative result first, theres a 1/255 change it's a false negative. To rule that out, we must retry again with possible_first_byte set to false
            if padding_oracle_result is None:
                self.debug("still could be in a possible_first_byte situation - retrying with different first byte")
                padding_oracle_result = await self.padding_oracle_execute(
                    data, encoding, block_size, cookies, possible_first_byte=False
                )

            if padding_oracle_result is True:
                context = f"Lightfuzz Cryptographic Probe Submodule detected a probable padding oracle vulnerability after manipulating parameter: [{self.event.data['name']}]"
                self.results.append(
                    {
                        "type": "VULNERABILITY",
                        "severity": "HIGH",
                        "description": f"Padding Oracle Vulnerability. Block size: [{str(block_size)}] {self.metadata()}",
                        "context": context,
                    }
                )

    async def error_string_search(self, text_dict, baseline_text):
        """
        Search for cryptographic error strings using YARA rules in the provided text dictionary and baseline text.
        """
        matching_techniques = set()
        matching_strings = set()

        # Check each manipulation technique
        for label, text in text_dict.items():
            matches = await self.lightfuzz.helpers.yara.match(self.compiled_rules, text)
            if matches:
                matching_techniques.add(label)
                for matched_string in matches:
                    matching_strings.add(matched_string)

        # Check for false positives by scanning baseline text
        context = f"Lightfuzz Cryptographic Probe Submodule detected a cryptographic error after manipulating parameter: [{self.event.data['name']}]"
        if matching_strings:
            baseline_matches = await self.lightfuzz.helpers.yara.match(self.compiled_rules, baseline_text)
            baseline_strings = set()
            for matched_string in baseline_matches:
                baseline_strings.add(matched_string)

            # Only report strings that weren't in the baseline
            unique_matches = matching_strings - baseline_strings
            if unique_matches:
                self.results.append(
                    {
                        "type": "FINDING",
                        "description": f"Possible Cryptographic Error. {self.metadata()} Strings: [{','.join(unique_matches)}] Detection Technique(s): [{','.join(matching_techniques)}]",
                        "context": context,
                    }
                )

            else:
                self.debug(
                    f"Aborting cryptographic error reporting - baseline_text already contained detected string(s) ({','.join(baseline_strings)})"
                )

    # Identify the hash function based on the length of the hash
    @staticmethod
    def identify_hash_function(hash_bytes):
        hash_length = len(hash_bytes)
        hash_functions = {
            16: hashlib.md5,
            20: hashlib.sha1,
            32: hashlib.sha256,
            48: hashlib.sha384,
            64: hashlib.sha512,
        }

        if hash_length in hash_functions:
            return hash_functions[hash_length]

    async def fuzz(self):
        cookies = self.event.data.get("assigned_cookies", {})
        probe_value = self.incoming_probe_value(populate_empty=False)

        if not probe_value:
            self.debug(
                f"The Cryptography Probe Submodule requires original value, aborting [{self.event.data['type']}] [{self.event.data['name']}]"
            )
            return

        # obtain the baseline probe to compare against
        baseline_probe = await self.baseline_probe(cookies)
        if not baseline_probe:
            self.verbose(f"Couldn't get baseline_probe for url {self.event.data['url']}, aborting")
            return

        # perform the manipulation techniques
        try:
            truncate_probe_value = self.modify_string(probe_value, action="truncate")
            mutate_probe_value = self.modify_string(probe_value, action="mutate")
        except ValueError as e:
            self.debug(f"Encountered error modifying value for parameter [{self.event.data['name']}]: {e} , aborting")
            return

        # Basic crypanalysis
        likely_crypto, possible_block_cipher = self.cryptanalysis(probe_value)

        # if the value is not likely to be cryptographic, we can skip the rest of the tests
        if not likely_crypto:
            self.debug("Parameter value does not appear to be cryptographic, aborting tests")
            return

        # Cryptographic Response Divergence Test

        http_compare = self.compare_baseline(self.event.data["type"], probe_value, cookies)
        try:
            arbitrary_probe = await self.compare_probe(http_compare, self.event.data["type"], "AAAAAAA", cookies)  #
            truncate_probe = await self.compare_probe(
                http_compare, self.event.data["type"], truncate_probe_value, cookies
            )  # manipulate the value by truncating a byte
            mutate_probe = await self.compare_probe(
                http_compare, self.event.data["type"], mutate_probe_value, cookies
            )  # manipulate the value by mutating a byte in place
        except HttpCompareError as e:
            self.verbose(f"Encountered HttpCompareError Sending Compare Probe: {e}")
            return

        confirmed_techniques = []
        # mutate_probe[0] will be false if the response is different - mutate_probe[1] stores what aspect of the response is different (headers, body, code)
        # ensure the difference is in the body and not the headers or code
        # if the body is different and not empty, we have confirmed that single-byte mutation affected the response body
        if mutate_probe[0] is False and "body" in mutate_probe[1]:
            if (http_compare.compare_body(mutate_probe[3].text, arbitrary_probe[3].text) is False) or mutate_probe[
                3
            ].text == "":
                confirmed_techniques.append("Single-byte Mutation")

        # if the body is different and not empty, we have confirmed that byte truncation affected the response body
        if truncate_probe[0] is False and "body" in truncate_probe[1]:
            if (http_compare.compare_body(truncate_probe[3].text, arbitrary_probe[3].text) is False) or truncate_probe[
                3
            ].text == "":
                confirmed_techniques.append("Data Truncation")

        if confirmed_techniques:
            context = f"Lightfuzz Cryptographic Probe Submodule detected a parameter ({self.event.data['name']}) to appears to drive a cryptographic operation"
            self.results.append(
                {
                    "type": "FINDING",
                    "description": f"Probable Cryptographic Parameter. {self.metadata()} Detection Technique(s): [{', '.join(confirmed_techniques)}]",
                    "context": context,
                }
            )

        # Cryptographic Error String Test
        # Check if cryptographic error strings are present in the response after performing the manipulation techniques
        await self.error_string_search(
            {"truncate value": truncate_probe[3].text, "mutate value": mutate_probe[3].text}, baseline_probe.text
        )
        # if we have any confirmed techniques, or the word "padding" is in the response, we need to check for a padding oracle
        if confirmed_techniques or (
            "padding" in truncate_probe[3].text.lower() or "padding" in mutate_probe[3].text.lower()
        ):
            # Padding Oracle Test
            if possible_block_cipher:
                self.debug(
                    "Attempting padding oracle exploit since it looks like a block cipher and we have confirmed crypto"
                )
                await self.padding_oracle(probe_value, cookies)

            # Hash identification / Potential Length extension attack
            data, encoding = crypto.format_agnostic_decode(probe_value)
            # see if its possible that a given value is a hash, and if so, which one
            hash_function = self.identify_hash_function(data)
            if hash_function:
                hash_instance = hash_function()
                # if there are any hash functions which match the length, we check the additional parameters to see if they cause identical changes
                # this would indicate they are being used to generate the hash
                if (
                    hash_function
                    and "additional_params" in self.event.data.keys()
                    and self.event.data["additional_params"]
                ):
                    # for each additional parameter, we send a probe and check if it causes the same change in the response as the original probe
                    for additional_param_name, additional_param_value in self.event.data["additional_params"].items():
                        try:
                            additional_param_probe = await self.compare_probe(
                                http_compare,
                                self.event.data["type"],
                                probe_value,
                                cookies,
                                additional_params_override={additional_param_name: additional_param_value + "A"},
                            )
                        except HttpCompareError as e:
                            self.verbose(f"Encountered HttpCompareError Sending Compare Probe: {e}")
                            continue
                        # the additional parameter affects the potential hash parameter (suggesting its calculated in the hash)
                        # This is a potential length extension attack
                        if additional_param_probe[0] is False and (additional_param_probe[1] == mutate_probe[1]):
                            context = f"Lightfuzz Cryptographic Probe Submodule detected a parameter ({self.event.data['name']}) that is a likely a hash, which is connected to another parameter {additional_param_name})"
                            self.results.append(
                                {
                                    "type": "FINDING",
                                    "description": f"Possible {self.event.data['type']} parameter with {hash_instance.name.upper()} Hash as value. {self.metadata()}, linked to additional parameter [{additional_param_name}]",
                                    "context": context,
                                }
                            )



from .base import BaseLightfuzz
from bbot.errors import HttpCompareError


class serial(BaseLightfuzz):
    """Finds parameters where serialized objects might be being deserialized.
    It starts by performing a baseline with a specially-crafted non-serialized payload, separated by type (base64, hex, php raw).
    This is designed to coax out an error that's not related to the decoding process.

    After performing the baseline (Which by design may contain an error), we check for two possible deserialization cases:

        1) Replacing the payload with a serialized object changes the status code to 200 (minus some string signatures to help prevent false positives)

        2) If the first case doesn't match, we check for a telltale error string like "java.io.optionaldataexception" in the response.
    """

    friendly_name = "Unsafe Deserialization"

    # Class-level constants
    CONTROL_PAYLOAD_HEX = "f56124208220432ec767646acd2e6c6bc9622a62c5656f2eeb616e2f"
    CONTROL_PAYLOAD_BASE64 = "4Wt5fYx5Y3rELn5myS5oa996Ji7IZ28uwGdha4x6YmuMfG992CA="
    CONTROL_PAYLOAD_PHP_RAW = "z:0:{}"

    BASE64_SERIALIZATION_PAYLOADS = {
        "php_base64": "YToxOntpOjA7aToxO30=",
        "java_base64": "rO0ABXNyABFqYXZhLmxhbmcuQm9vbGVhbs0gcoDVnPruAgABWgAFdmFsdWV4cAA=",
        "java_base64_string_error": "rO0ABXQABHRlc3Q=",
        "java_base64_OptionalDataException": "rO0ABXcEAAAAAAEAAAABc3IAEGphdmEudXRpbC5IYXNoTWFwAAAAAAAAAAECAAJMAARrZXkxYgABAAAAAAAAAAJ4cHcBAAAAB3QABHRlc3Q=",
        "dotnet_base64": "AAEAAAD/////AQAAAAAAAAAGAQAAAAdndXN0YXZvCw==",
        "ruby_base64": "BAh7BjoKbE1FAAVJsg==",
    }

    HEX_SERIALIZATION_PAYLOADS = {
        "java_hex": "ACED00057372000E6A6176612E6C616E672E426F6F6C65616ECD207EC0D59CF6EE02000157000576616C7565787000",
        "java_hex_OptionalDataException": "ACED0005737200106A6176612E7574696C2E486173684D617000000000000000012000014C00046B6579317A00010000000000000278707000000774000474657374",
        "dotnet_hex": "0001000000ffffffff01000000000000000601000000076775737461766f0b",
    }

    PHP_RAW_SERIALIZATION_PAYLOADS = {
        "php_raw": "a:0:{}",
    }

    SERIALIZATION_ERRORS = [
        "invalid user",
        "cannot cast java.lang.string",
        "dump format error",
        "java.io.optionaldataexception",
    ]

    GENERAL_ERRORS = [
        "Internal Error",
        "Internal Server Error",
        "The requested URL was rejected",
    ]

    def is_possibly_serialized(self, value):
        # Use the is_base64 method from BaseLightfuzz via self
        if self.is_base64(value):
            return True

        # Use the is_hex method from BaseLightfuzz via self
        if self.is_hex(value):
            return True

        # List of common PHP serialized data prefixes
        php_serialized_prefixes = [
            "a:",  # Array
            "O:",  # Object
            "s:",  # String
            "i:",  # Integer
            "d:",  # Double
            "b:",  # Boolean
            "N;",  # Null
        ]

        # Check if the value starts with any of the PHP serialized prefixes
        if any(value.startswith(prefix) for prefix in php_serialized_prefixes):
            return True
        return False

    async def fuzz(self):
        cookies = self.event.data.get("assigned_cookies", {})
        control_payload_hex = self.CONTROL_PAYLOAD_HEX
        control_payload_base64 = self.CONTROL_PAYLOAD_BASE64
        control_payload_php_raw = self.CONTROL_PAYLOAD_PHP_RAW

        base64_serialization_payloads = self.BASE64_SERIALIZATION_PAYLOADS
        hex_serialization_payloads = self.HEX_SERIALIZATION_PAYLOADS
        php_raw_serialization_payloads = self.PHP_RAW_SERIALIZATION_PAYLOADS

        serialization_errors = self.SERIALIZATION_ERRORS
        general_errors = self.GENERAL_ERRORS

        probe_value = self.incoming_probe_value(populate_empty=False)
        if probe_value:
            if self.is_possibly_serialized(probe_value):
                self.debug(
                    f"Existing value is not ruled out for being a serialized object, proceeding [{self.event.data['type']}] [{self.event.data['name']}]"
                )
            else:
                self.debug(
                    f"The Serialization Submodule only operates when there is no original value, or when the original value could potentially be a serialized object, aborting [{self.event.data['type']}] [{self.event.data['name']}]"
                )
                return

        try:
            http_compare_hex = self.compare_baseline(self.event.data["type"], control_payload_hex, cookies)
            http_compare_base64 = self.compare_baseline(self.event.data["type"], control_payload_base64, cookies)
            http_compare_php_raw = self.compare_baseline(self.event.data["type"], control_payload_php_raw, cookies)
        except HttpCompareError as e:
            self.debug(f"HttpCompareError encountered: {e}")
            return

        # Proceed with payload probes
        for payload_set, payload_baseline in [
            (base64_serialization_payloads, http_compare_base64),
            (hex_serialization_payloads, http_compare_hex),
            (php_raw_serialization_payloads, http_compare_php_raw),
        ]:
            for type, payload in payload_set.items():
                try:
                    matches_baseline, diff_reasons, reflection, response = await self.compare_probe(
                        payload_baseline, self.event.data["type"], payload, cookies
                    )
                except HttpCompareError as e:
                    self.debug(f"HttpCompareError encountered: {e}")
                    continue

                if matches_baseline:
                    self.debug(f"Payload {type} matches baseline, skipping")
                    continue

                self.debug(f"Probe result for {type}: {response}")

                status_code = getattr(response, "status_code", 0)
                if status_code == 0:
                    continue

                if diff_reasons == ["header"]:
                    self.debug(f"Only header diffs found for {type}, skipping")
                    continue

                if status_code not in (200, 500):
                    self.debug(f"Status code {status_code} not in (200, 500), skipping")
                    continue

                # if the status code changed to 200, and the response doesn't match our general error exclusions, we have a finding
                self.debug(f"Potential finding detected for {type}, needs confirmation")
                if (
                    status_code == 200
                    and "code" in diff_reasons
                    and not any(
                        error in response.text for error in general_errors
                    )  # ensure the 200 is not actually an error
                ):

                    def get_title(text):
                        soup = self.lightfuzz.helpers.beautifulsoup(text, "html.parser")
                        if soup and soup.title and soup.title.string:
                            return f"'{self.lightfuzz.helpers.truncate_string(soup.title.string, 50)}'"
                        return ""

                    baseline_title = get_title(payload_baseline.baseline.text)
                    probe_title = get_title(response.text)

                    self.results.append(
                        {
                            "type": "FINDING",
                            "description": f"POSSIBLE Unsafe Deserialization. {self.metadata()} Technique: [Error Resolution (Baseline: [{payload_baseline.baseline.status_code}] {baseline_title} -> Probe: [{status_code}] {probe_title})] Serialization Payload: [{type}]",
                        }
                    )
                # if the first case doesn't match, we check for a telltale error string like "java.io.optionaldataexception" in the response.
                # but only if the response is a 500, or a 200 with a body diff
                elif status_code == 500 or (status_code == 200 and diff_reasons == ["body"]):
                    self.debug(f"500 status code or body match for {type}")
                    for serialization_error in serialization_errors:
                        # check for the error string, but also ensure the error string isn't just always present in the response
                        if (
                            serialization_error in response.text.lower()
                            and serialization_error not in payload_baseline.baseline.text.lower()
                        ):
                            self.debug(f"Error string '{serialization_error}' found in response for {type}")
                            self.results.append(
                                {
                                    "type": "FINDING",
                                    "description": f"POSSIBLE Unsafe Deserialization. {self.metadata()} Technique: [Differential Error Analysis] Error-String: [{serialization_error}] Payload: [{type}]",
                                }
                            )
                            break



from .base import ModuleTestBase, tempwordlist


class TestFFUFShortnames(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    test_wordlist = ["11111111", "administrator", "portal", "console", "junkword1", "zzzjunkword2", "directory"]
    config_overrides = {
        "modules": {
            "ffuf_shortnames": {
                "find_common_prefixes": True,
                "find_subwords": True,
                "wordlist": tempwordlist(test_wordlist),
            }
        }
    }
    modules_overrides = ["ffuf_shortnames", "httpx"]

    async def setup_after_prep(self, module_test):
        module_test.httpserver.no_handler_status_code = 404

        seed_events = []
        parent_event = module_test.scan.make_event(
            "http://127.0.0.1:8888/",
            "URL",
            module_test.scan.root_event,
            module="httpx",
            tags=["status-200", "distance-0"],
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ADMINI~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ADM_PO~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ABCZZZ~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ABCXXX~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ABCYYY~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ABCCON~1.ASP",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/DIRECT~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/ADM_DI~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/XYZDIR~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/XYZAAA~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/XYZBBB~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/XYZCCC~1",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-directory"],
            )
        )
        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/SHORT~1.PL",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )

        seed_events.append(
            module_test.scan.make_event(
                "http://127.0.0.1:8888/newpro~1.asp",
                "URL_HINT",
                parent_event,
                module="iis_shortnames",
                tags=["shortname-endpoint"],
            )
        )
        for event in seed_events:
            await module_test.scan.ingress_module.incoming_event_queue.put(event)

        expect_args = {"method": "GET", "uri": "/administrator.aspx"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/adm_portal.aspx"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/abcconsole.aspx"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/directory/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/adm_directory/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/xyzdirectory/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/short.pl"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/newproxy.aspx"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        basic_detection = False
        directory_detection = False
        prefix_detection = False
        delimiter_detection = False
        directory_delimiter_detection = False
        prefix_delimiter_detection = False
        short_extensions_detection = False
        subword_detection = False

        for e in events:
            if e.type == "URL_UNVERIFIED":
                if e.data == "http://127.0.0.1:8888/administrator.aspx":
                    basic_detection = True
                if e.data == "http://127.0.0.1:8888/directory/":
                    directory_detection = True
                if e.data == "http://127.0.0.1:8888/adm_portal.aspx":
                    prefix_detection = True
                if e.data == "http://127.0.0.1:8888/abcconsole.aspx":
                    delimiter_detection = True
                if e.data == "http://127.0.0.1:8888/adm_directory/":
                    directory_delimiter_detection = True
                if e.data == "http://127.0.0.1:8888/xyzdirectory/":
                    prefix_delimiter_detection = True
                if e.data == "http://127.0.0.1:8888/short.pl":
                    short_extensions_detection = True
                if e.data == "http://127.0.0.1:8888/newproxy.aspx":
                    subword_detection = True

        assert basic_detection
        assert directory_detection
        assert prefix_detection
        assert delimiter_detection
        assert directory_delimiter_detection
        assert prefix_delimiter_detection
        assert short_extensions_detection
        assert subword_detection



import httpx

from .base import ModuleTestBase


class TestRapidDNS(ModuleTestBase):
    web_body = """<th scope="row ">12</th>
<td>asdf.blacklanternsecurity.com</td>
<td><a href="/sameip/asdf.blacklanternsecurity.com.?t=cname#result" target="_blank" title="asdf.blacklanternsecurity.com. same ip website">asdf.blacklanternsecurity.com.</a>"""

    async def setup_after_prep(self, module_test):
        module_test.module.abort_if = lambda e: False
        module_test.httpx_mock.add_response(
            url="https://rapiddns.io/subdomain/blacklanternsecurity.com?full=1#result", text=self.web_body
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"


class TestRapidDNSAbortThreshold1(TestRapidDNS):
    module_name = "rapiddns"

    async def setup_after_prep(self, module_test):
        self.url_count = {}

        async def custom_callback(request):
            url = str(request.url)
            try:
                self.url_count[url] += 1
            except KeyError:
                self.url_count[url] = 1
            raise httpx.TimeoutException("timeout")

        module_test.httpx_mock.add_callback(custom_callback)

        await module_test.mock_dns(
            {
                "blacklanternsecurity.com": {"A": ["127.0.0.88"]},
                "evilcorp.com": {"A": ["127.0.0.11"]},
                "evilcorp.net": {"A": ["127.0.0.22"]},
                "evilcorp.co.uk": {"A": ["127.0.0.33"]},
            }
        )

    def check(self, module_test, events):
        assert module_test.module.api_failure_abort_threshold == 10
        assert module_test.module.errored is False
        assert module_test.module._api_request_failures == 3
        assert module_test.module.api_retries == 3
        assert {e.data for e in events if e.type == "DNS_NAME"} == {"blacklanternsecurity.com"}
        assert self.url_count == {
            "https://rapiddns.io/subdomain/blacklanternsecurity.com?full=1#result": 3,
        }


class TestRapidDNSAbortThreshold2(TestRapidDNSAbortThreshold1):
    targets = ["blacklanternsecurity.com", "evilcorp.com"]

    def check(self, module_test, events):
        assert module_test.module.api_failure_abort_threshold == 10
        assert module_test.module.errored is False
        assert module_test.module._api_request_failures == 6
        assert module_test.module.api_retries == 3
        assert {e.data for e in events if e.type == "DNS_NAME"} == {"blacklanternsecurity.com", "evilcorp.com"}
        assert self.url_count == {
            "https://rapiddns.io/subdomain/blacklanternsecurity.com?full=1#result": 3,
            "https://rapiddns.io/subdomain/evilcorp.com?full=1#result": 3,
        }


class TestRapidDNSAbortThreshold3(TestRapidDNSAbortThreshold1):
    targets = ["blacklanternsecurity.com", "evilcorp.com", "evilcorp.net"]

    def check(self, module_test, events):
        assert module_test.module.api_failure_abort_threshold == 10
        assert module_test.module.errored is False
        assert module_test.module._api_request_failures == 9
        assert module_test.module.api_retries == 3
        assert {e.data for e in events if e.type == "DNS_NAME"} == {
            "blacklanternsecurity.com",
            "evilcorp.com",
            "evilcorp.net",
        }
        assert self.url_count == {
            "https://rapiddns.io/subdomain/blacklanternsecurity.com?full=1#result": 3,
            "https://rapiddns.io/subdomain/evilcorp.com?full=1#result": 3,
            "https://rapiddns.io/subdomain/evilcorp.net?full=1#result": 3,
        }


class TestRapidDNSAbortThreshold4(TestRapidDNSAbortThreshold1):
    targets = ["blacklanternsecurity.com", "evilcorp.com", "evilcorp.net", "evilcorp.co.uk"]

    def check(self, module_test, events):
        assert module_test.module.api_failure_abort_threshold == 10
        assert module_test.module.errored is True
        assert module_test.module._api_request_failures == 10
        assert module_test.module.api_retries == 3
        assert {e.data for e in events if e.type == "DNS_NAME"} == {
            "blacklanternsecurity.com",
            "evilcorp.com",
            "evilcorp.net",
            "evilcorp.co.uk",
        }
        assert len(self.url_count) == 4
        assert list(self.url_count.values()).count(3) == 3
        assert list(self.url_count.values()).count(1) == 1



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class securitytrails(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query the SecurityTrails API for subdomains",
        "created_date": "2022-07-03",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "SecurityTrails API key"}

    base_url = "https://api.securitytrails.com/v1"
    ping_url = f"{base_url}/ping?apikey={{api_key}}"

    async def setup(self):
        self.limit = 100
        return await super().setup()

    async def request_url(self, query):
        url = f"{self.base_url}/domain/{query}/subdomains?apikey={{api_key}}"
        response = await self.api_request(url)
        return response

    async def parse_results(self, r, query):
        results = set()
        j = r.json()
        if isinstance(j, dict):
            for host in j.get("subdomains", []):
                results.add(f"{host}.{query}")
        return results



import pytest


async def test_web_envelopes():
    from bbot.core.helpers.web.envelopes import (
        BaseEnvelope,
        TextEnvelope,
        HexEnvelope,
        B64Envelope,
        JSONEnvelope,
        XMLEnvelope,
        URLEnvelope,
    )

    # simple text
    text_envelope = BaseEnvelope.detect("foo")
    assert isinstance(text_envelope, TextEnvelope)
    assert text_envelope.unpacked_data() == "foo"
    assert text_envelope.subparams == {"__default__": "foo"}
    expected_subparams = [([], "foo")]
    assert list(text_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert text_envelope.get_subparam(subparam) == value
    assert text_envelope.pack() == "foo"
    assert text_envelope.num_envelopes == 0
    assert text_envelope.get_subparam() == "foo"
    text_envelope.set_subparam(value="bar")
    assert text_envelope.get_subparam() == "bar"
    assert text_envelope.unpacked_data() == "bar"

    # simple binary
    # binary_envelope = BaseEnvelope.detect("foo\x00")
    # assert isinstance(binary_envelope, BinaryEnvelope)
    # assert binary_envelope.unpacked_data == "foo\x00"
    # assert binary_envelope.packed_data == "foo\x00"
    # assert binary_envelope.subparams == {"__default__": "foo\x00"}

    # text encoded as hex
    hex_envelope = BaseEnvelope.detect("706172616d")
    assert isinstance(hex_envelope, HexEnvelope)
    assert hex_envelope.unpacked_data(recursive=True) == "param"
    hex_inner_envelope = hex_envelope.unpacked_data(recursive=False)
    assert isinstance(hex_inner_envelope, TextEnvelope)
    assert hex_inner_envelope.unpacked_data(recursive=False) == "param"
    assert hex_inner_envelope.unpacked_data(recursive=True) == "param"
    assert list(hex_envelope.get_subparams(recursive=False)) == [([], hex_inner_envelope)]
    assert list(hex_envelope.get_subparams(recursive=True)) == [([], "param")]
    assert hex_inner_envelope.unpacked_data() == "param"
    assert hex_inner_envelope.subparams == {"__default__": "param"}
    expected_subparams = [([], "param")]
    assert list(hex_inner_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert hex_inner_envelope.get_subparam(subparam) == value
    assert hex_envelope.pack() == "706172616d"
    assert hex_envelope.num_envelopes == 1
    assert hex_envelope.get_subparam() == "param"
    hex_envelope.set_subparam(value="asdf")
    assert hex_envelope.get_subparam() == "asdf"
    assert hex_envelope.unpacked_data() == "asdf"
    assert hex_envelope.pack() == "61736466"

    # text encoded as base64
    base64_envelope = BaseEnvelope.detect("cGFyYW0=")
    assert isinstance(base64_envelope, B64Envelope)
    assert base64_envelope.unpacked_data() == "param"
    base64_inner_envelope = base64_envelope.unpacked_data(recursive=False)
    assert isinstance(base64_inner_envelope, TextEnvelope)
    assert list(base64_envelope.get_subparams(recursive=False)) == [([], base64_inner_envelope)]
    assert list(base64_envelope.get_subparams()) == [([], "param")]
    assert base64_inner_envelope.pack() == "param"
    assert base64_inner_envelope.unpacked_data() == "param"
    assert base64_inner_envelope.subparams == {"__default__": "param"}
    expected_subparams = [([], "param")]
    assert list(base64_inner_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert base64_inner_envelope.get_subparam(subparam) == value
    assert base64_envelope.num_envelopes == 1
    base64_envelope.set_subparam(value="asdf")
    assert base64_envelope.get_subparam() == "asdf"
    assert base64_envelope.unpacked_data() == "asdf"
    assert base64_envelope.pack() == "YXNkZg=="

    # test inside hex inside base64
    hex_envelope = BaseEnvelope.detect("634746795957303d")
    assert isinstance(hex_envelope, HexEnvelope)
    assert hex_envelope.get_subparam() == "param"
    assert hex_envelope.unpacked_data() == "param"
    base64_envelope = hex_envelope.unpacked_data(recursive=False)
    assert isinstance(base64_envelope, B64Envelope)
    assert base64_envelope.get_subparam() == "param"
    assert base64_envelope.unpacked_data() == "param"
    text_envelope = base64_envelope.unpacked_data(recursive=False)
    assert isinstance(text_envelope, TextEnvelope)
    assert text_envelope.get_subparam() == "param"
    assert text_envelope.unpacked_data() == "param"
    hex_envelope.set_subparam(value="asdf")
    assert hex_envelope.get_subparam() == "asdf"
    assert hex_envelope.unpacked_data() == "asdf"
    assert text_envelope.get_subparam() == "asdf"
    assert text_envelope.unpacked_data() == "asdf"
    assert base64_envelope.get_subparam() == "asdf"
    assert base64_envelope.unpacked_data() == "asdf"

    # URL-encoded text
    url_encoded_envelope = BaseEnvelope.detect("a%20b%20c")
    assert isinstance(url_encoded_envelope, URLEnvelope)
    assert url_encoded_envelope.pack() == "a%20b%20c"
    assert url_encoded_envelope.unpacked_data() == "a b c"
    url_inner_envelope = url_encoded_envelope.unpacked_data(recursive=False)
    assert isinstance(url_inner_envelope, TextEnvelope)
    assert url_inner_envelope.unpacked_data(recursive=False) == "a b c"
    assert url_inner_envelope.unpacked_data(recursive=True) == "a b c"
    assert list(url_encoded_envelope.get_subparams(recursive=False)) == [([], url_inner_envelope)]
    assert list(url_encoded_envelope.get_subparams(recursive=True)) == [([], "a b c")]
    assert url_inner_envelope.pack() == "a b c"
    assert url_inner_envelope.unpacked_data() == "a b c"
    assert url_inner_envelope.subparams == {"__default__": "a b c"}
    expected_subparams = [([], "a b c")]
    assert list(url_inner_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert url_inner_envelope.get_subparam(subparam) == value
    assert url_encoded_envelope.num_envelopes == 1
    url_encoded_envelope.set_subparam(value="a s d f")
    assert url_encoded_envelope.get_subparam() == "a s d f"
    assert url_encoded_envelope.unpacked_data() == "a s d f"
    assert url_encoded_envelope.pack() == "a%20s%20d%20f"

    # json
    json_envelope = BaseEnvelope.detect('{"param1": "val1", "param2": {"param3": "val3"}}')
    assert isinstance(json_envelope, JSONEnvelope)
    assert json_envelope.pack() == '{"param1": "val1", "param2": {"param3": "val3"}}'
    assert json_envelope.unpacked_data() == {"param1": "val1", "param2": {"param3": "val3"}}
    assert json_envelope.unpacked_data(recursive=False) == {"param1": "val1", "param2": {"param3": "val3"}}
    assert json_envelope.unpacked_data(recursive=True) == {"param1": "val1", "param2": {"param3": "val3"}}
    assert json_envelope.subparams == {"param1": "val1", "param2": {"param3": "val3"}}
    expected_subparams = [
        (["param1"], "val1"),
        (["param2", "param3"], "val3"),
    ]
    assert list(json_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert json_envelope.get_subparam(subparam) == value
    json_envelope.selected_subparam = ["param2", "param3"]
    assert json_envelope.get_subparam() == "val3"
    assert json_envelope.num_envelopes == 1

    # prevent json over-detection
    just_a_string = BaseEnvelope.detect("10")
    assert not isinstance(just_a_string, JSONEnvelope)

    # xml
    xml_envelope = BaseEnvelope.detect(
        '<root><param1 attr="attr1">val1</param1><param2><param3>val3</param3></param2></root>'
    )
    assert isinstance(xml_envelope, XMLEnvelope)
    assert (
        xml_envelope.pack()
        == '<?xml version="1.0" encoding="utf-8"?>\n<root><param1 attr="attr1">val1</param1><param2><param3>val3</param3></param2></root>'
    )
    assert xml_envelope.unpacked_data() == {
        "root": {"param1": {"@attr": "attr1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    assert xml_envelope.unpacked_data(recursive=False) == {
        "root": {"param1": {"@attr": "attr1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    assert xml_envelope.unpacked_data(recursive=True) == {
        "root": {"param1": {"@attr": "attr1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    assert xml_envelope.subparams == {
        "root": {"param1": {"@attr": "attr1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    expected_subparams = [
        (["root", "param1", "@attr"], "attr1"),
        (["root", "param1", "#text"], "val1"),
        (["root", "param2", "param3"], "val3"),
    ]
    assert list(xml_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert xml_envelope.get_subparam(subparam) == value
    assert xml_envelope.num_envelopes == 1

    # json inside base64
    base64_json_envelope = BaseEnvelope.detect("eyJwYXJhbTEiOiAidmFsMSIsICJwYXJhbTIiOiB7InBhcmFtMyI6ICJ2YWwzIn19")
    assert isinstance(base64_json_envelope, B64Envelope)
    assert base64_json_envelope.pack() == "eyJwYXJhbTEiOiAidmFsMSIsICJwYXJhbTIiOiB7InBhcmFtMyI6ICJ2YWwzIn19"
    assert base64_json_envelope.unpacked_data() == {"param1": "val1", "param2": {"param3": "val3"}}
    base64_inner_envelope = base64_json_envelope.unpacked_data(recursive=False)
    assert isinstance(base64_inner_envelope, JSONEnvelope)
    assert base64_inner_envelope.pack() == '{"param1": "val1", "param2": {"param3": "val3"}}'
    assert base64_inner_envelope.unpacked_data() == {"param1": "val1", "param2": {"param3": "val3"}}
    assert base64_inner_envelope.subparams == {"param1": "val1", "param2": {"param3": "val3"}}
    expected_subparams = [
        (["param1"], "val1"),
        (["param2", "param3"], "val3"),
    ]
    assert list(base64_json_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert base64_json_envelope.get_subparam(subparam) == value
    assert base64_json_envelope.num_envelopes == 2
    with pytest.raises(ValueError):
        assert base64_json_envelope.get_subparam()
    base64_json_envelope.selected_subparam = ["param2", "param3"]
    assert base64_json_envelope.get_subparam() == "val3"

    # xml inside url inside hex inside base64
    nested_xml_envelope = BaseEnvelope.detect(
        "MjUzMzYzMjUzNzMyMjUzNjY2MjUzNjY2MjUzNzM0MjUzMzY1MjUzMzYzMjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMxMjUzMjMwMjUzNjMxMjUzNzM0MjUzNzM0MjUzNzMyMjUzMzY0MjUzMjMyMjUzNzM2MjUzNjMxMjUzNjYzMjUzMzMxMjUzMjMyMjUzMzY1MjUzNzM2MjUzNjMxMjUzNjYzMjUzMzMxMjUzMzYzMjUzMjY2MjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMxMjUzMzY1MjUzMzYzMjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMyMjUzMzY1MjUzMzYzMjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMzMjUzMzY1MjUzNzM2MjUzNjMxMjUzNjYzMjUzMzMzMjUzMzYzMjUzMjY2MjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMzMjUzMzY1MjUzMzYzMjUzMjY2MjUzNzMwMjUzNjMxMjUzNzMyMjUzNjMxMjUzNjY0MjUzMzMyMjUzMzY1MjUzMzYzMjUzMjY2MjUzNzMyMjUzNjY2MjUzNjY2MjUzNzM0MjUzMzY1"
    )
    assert isinstance(nested_xml_envelope, B64Envelope)
    assert nested_xml_envelope.unpacked_data() == {
        "root": {"param1": {"@attr": "val1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    assert (
        nested_xml_envelope.pack()
        == "MjUzMzQzMjUzMzQ2Nzg2ZDZjMjUzMjMwNzY2NTcyNzM2OTZmNmUyNTMzNDQyNTMyMzIzMTJlMzAyNTMyMzIyNTMyMzA2NTZlNjM2ZjY0Njk2ZTY3MjUzMzQ0MjUzMjMyNzU3NDY2MmQzODI1MzIzMjI1MzM0NjI1MzM0NTI1MzA0MTI1MzM0MzcyNmY2Zjc0MjUzMzQ1MjUzMzQzNzA2MTcyNjE2ZDMxMjUzMjMwNjE3NDc0NzIyNTMzNDQyNTMyMzI3NjYxNmMzMTI1MzIzMjI1MzM0NTc2NjE2YzMxMjUzMzQzMmY3MDYxNzI2MTZkMzEyNTMzNDUyNTMzNDM3MDYxNzI2MTZkMzIyNTMzNDUyNTMzNDM3MDYxNzI2MTZkMzMyNTMzNDU3NjYxNmMzMzI1MzM0MzJmNzA2MTcyNjE2ZDMzMjUzMzQ1MjUzMzQzMmY3MDYxNzI2MTZkMzIyNTMzNDUyNTMzNDMyZjcyNmY2Zjc0MjUzMzQ1"
    )
    inner_hex_envelope = nested_xml_envelope.unpacked_data(recursive=False)
    assert isinstance(inner_hex_envelope, HexEnvelope)
    assert (
        inner_hex_envelope.pack()
        == "253343253346786d6c25323076657273696f6e253344253232312e30253232253230656e636f64696e672533442532327574662d38253232253346253345253041253343726f6f74253345253343706172616d312532306174747225334425323276616c3125323225334576616c312533432f706172616d31253345253343706172616d32253345253343706172616d3325334576616c332533432f706172616d332533452533432f706172616d322533452533432f726f6f74253345"
    )
    inner_url_envelope = inner_hex_envelope.unpacked_data(recursive=False)
    assert isinstance(inner_url_envelope, URLEnvelope)
    assert (
        inner_url_envelope.pack()
        == r"%3C%3Fxml%20version%3D%221.0%22%20encoding%3D%22utf-8%22%3F%3E%0A%3Croot%3E%3Cparam1%20attr%3D%22val1%22%3Eval1%3C/param1%3E%3Cparam2%3E%3Cparam3%3Eval3%3C/param3%3E%3C/param2%3E%3C/root%3E"
    )
    inner_xml_envelope = inner_url_envelope.unpacked_data(recursive=False)
    assert isinstance(inner_xml_envelope, XMLEnvelope)
    assert (
        inner_xml_envelope.pack()
        == '<?xml version="1.0" encoding="utf-8"?>\n<root><param1 attr="val1">val1</param1><param2><param3>val3</param3></param2></root>'
    )
    assert inner_xml_envelope.unpacked_data() == {
        "root": {"param1": {"@attr": "val1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    assert inner_xml_envelope.subparams == {
        "root": {"param1": {"@attr": "val1", "#text": "val1"}, "param2": {"param3": "val3"}}
    }
    expected_subparams = [
        (["root", "param1", "@attr"], "val1"),
        (["root", "param1", "#text"], "val1"),
        (["root", "param2", "param3"], "val3"),
    ]
    assert list(nested_xml_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert nested_xml_envelope.get_subparam(subparam) == value
    assert nested_xml_envelope.num_envelopes == 4

    # manipulating text inside hex
    hex_envelope = BaseEnvelope.detect("706172616d")
    expected_subparams = [([], "param")]
    assert list(hex_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert hex_envelope.get_subparam(subparam) == value
    hex_envelope.set_subparam([], "asdf")
    expected_subparams = [([], "asdf")]
    assert list(hex_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert hex_envelope.get_subparam(subparam) == value
    assert hex_envelope.unpacked_data() == "asdf"

    # manipulating json inside base64
    base64_json_envelope = BaseEnvelope.detect("eyJwYXJhbTEiOiAidmFsMSIsICJwYXJhbTIiOiB7InBhcmFtMyI6ICJ2YWwzIn19")
    expected_subparams = [
        (["param1"], "val1"),
        (["param2", "param3"], "val3"),
    ]
    assert list(base64_json_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert base64_json_envelope.get_subparam(subparam) == value
    base64_json_envelope.set_subparam(["param1"], {"asdf": [None], "fdsa": 1.0})
    expected_subparams = [
        (["param1", "asdf"], [None]),
        (["param1", "fdsa"], 1.0),
        (["param2", "param3"], "val3"),
    ]
    assert list(base64_json_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert base64_json_envelope.get_subparam(subparam) == value
    base64_json_envelope.set_subparam(["param2", "param3"], {"1234": [None], "4321": 1.0})
    expected_subparams = [
        (["param1", "asdf"], [None]),
        (["param1", "fdsa"], 1.0),
        (["param2", "param3", "1234"], [None]),
        (["param2", "param3", "4321"], 1.0),
    ]
    assert list(base64_json_envelope.get_subparams()) == expected_subparams
    base64_json_envelope.set_subparam(["param2"], None)
    expected_subparams = [
        (["param1", "asdf"], [None]),
        (["param1", "fdsa"], 1.0),
        (["param2"], None),
    ]
    assert list(base64_json_envelope.get_subparams()) == expected_subparams

    # xml inside url inside base64
    xml_envelope = BaseEnvelope.detect(
        "JTNDP3htbCUyMHZlcnNpb249JTIyMS4wJTIyJTIwZW5jb2Rpbmc9JTIydXRmLTglMjI/JTNFJTBBJTNDcm9vdCUzRSUzQ3BhcmFtMSUyMGF0dHI9JTIydmFsMSUyMiUzRXZhbDElM0MvcGFyYW0xJTNFJTNDcGFyYW0yJTNFJTNDcGFyYW0zJTNFdmFsMyUzQy9wYXJhbTMlM0UlM0MvcGFyYW0yJTNFJTNDL3Jvb3QlM0U="
    )
    assert (
        xml_envelope.pack()
        == "JTNDJTNGeG1sJTIwdmVyc2lvbiUzRCUyMjEuMCUyMiUyMGVuY29kaW5nJTNEJTIydXRmLTglMjIlM0YlM0UlMEElM0Nyb290JTNFJTNDcGFyYW0xJTIwYXR0ciUzRCUyMnZhbDElMjIlM0V2YWwxJTNDL3BhcmFtMSUzRSUzQ3BhcmFtMiUzRSUzQ3BhcmFtMyUzRXZhbDMlM0MvcGFyYW0zJTNFJTNDL3BhcmFtMiUzRSUzQy9yb290JTNF"
    )
    expected_subparams = [
        (["root", "param1", "@attr"], "val1"),
        (["root", "param1", "#text"], "val1"),
        (["root", "param2", "param3"], "val3"),
    ]
    assert list(xml_envelope.get_subparams()) == expected_subparams
    xml_envelope.set_subparam(["root", "param1", "@attr"], "asdf")
    expected_subparams = [
        (["root", "param1", "@attr"], "asdf"),
        (["root", "param1", "#text"], "val1"),
        (["root", "param2", "param3"], "val3"),
    ]
    assert list(xml_envelope.get_subparams()) == expected_subparams
    assert (
        xml_envelope.pack()
        == "JTNDJTNGeG1sJTIwdmVyc2lvbiUzRCUyMjEuMCUyMiUyMGVuY29kaW5nJTNEJTIydXRmLTglMjIlM0YlM0UlMEElM0Nyb290JTNFJTNDcGFyYW0xJTIwYXR0ciUzRCUyMmFzZGYlMjIlM0V2YWwxJTNDL3BhcmFtMSUzRSUzQ3BhcmFtMiUzRSUzQ3BhcmFtMyUzRXZhbDMlM0MvcGFyYW0zJTNFJTNDL3BhcmFtMiUzRSUzQy9yb290JTNF"
    )
    xml_envelope.set_subparam(["root", "param2", "param3"], {"1234": [None], "4321": 1.0})
    expected_subparams = [
        (["root", "param1", "@attr"], "asdf"),
        (["root", "param1", "#text"], "val1"),
        (["root", "param2", "param3", "1234"], [None]),
        (["root", "param2", "param3", "4321"], 1.0),
    ]
    assert list(xml_envelope.get_subparams()) == expected_subparams

    # null
    null_envelope = BaseEnvelope.detect("null")
    assert isinstance(null_envelope, JSONEnvelope)
    assert null_envelope.unpacked_data() is None
    assert null_envelope.pack() == "null"
    expected_subparams = [([], None)]
    assert list(null_envelope.get_subparams()) == expected_subparams
    for subparam, value in expected_subparams:
        assert null_envelope.get_subparam(subparam) == value

    tiny_base64 = BaseEnvelope.detect("YWJi")
    assert isinstance(tiny_base64, TextEnvelope)
